/*
 * @(#)msd_dir.c 1.4 87/11/06	Public Domain.
 *
 *  A public domain implementation of BSD directory routines for
 *  MS-DOS.  Written by Michael Rendell ({uunet,utai}michael@garfield),
 *  August 1897
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <dirent.h>
#include <dos.h>

#define	ATTRIBUTES		(_A_SUBDIR | _A_HIDDEN | _A_SYSTEM | \
				 _A_NORMAL | _A_RDONLY | _A_ARCH | _A_VOLID)

static void	free_dircontents (struct _dircontents *);

DIR	*opendir(name)
char	*name;
{
    struct stat		statb;
    DIR			*dirp;
    char		c;
    struct _dircontents	*dp;
    char		nbuf[PATH_MAX + NAME_MAX + 2];
    struct find_t	dtabuf;
    
    if (stat (name, &statb) < 0)
	return (DIR *) NULL;

    if (!S_ISDIR(statb.st_mode))
    {
	errno = ENOTDIR;
	return (DIR *)NULL;
    }

    if ((dirp = (DIR *) malloc (sizeof (DIR))) == (DIR *) NULL)
	return (DIR *) NULL;

    if (*name && (c = name[strlen (name) - 1]) != '\\' && c != '/')
	(void) strcat (strcpy (nbuf, name), "\\*.*");

    else
	(void) strcat (strcpy (nbuf, name), "*.*");

    dirp->dd_loc      = 0;
    dirp->dd_cp       = (struct _dircontents *) NULL;
    dirp->dd_contents = (struct _dircontents *) NULL;

    if (_dos_findfirst (nbuf, ATTRIBUTES, &dtabuf) != 0)
	return dirp;

    do 
    {
	if (((dp = (struct _dircontents *) malloc(sizeof(struct _dircontents))) == (struct _dircontents *) NULL) ||
	    ((dp->_d_entry = strdup (dtabuf.name)) == (char *) NULL))
	{
	    if (dp != (char *)NULL)
		free ((char *) dp);

	    free_dircontents (dirp->dd_contents);
	    return (DIR *) NULL;
	}

	if (dirp->dd_contents)
	    dirp->dd_cp = dirp->dd_cp->_d_next = dp;

	else
	    dirp->dd_contents = dirp->dd_cp = dp;

	dp->_d_next = (struct _dircontents *) NULL;

    } while (_dos_findnext (&dtabuf) == 0);

    dirp->dd_cp = dirp->dd_contents;

    return dirp;
}

int	closedir(dirp)
DIR	*dirp;
{
    free_dircontents (dirp->dd_contents);
    free ((char *) dirp);
    return 0;
}

struct dirent	*readdir(dirp)
DIR		*dirp;
{
    static struct dirent	dp;
    
    if (dirp->dd_cp == (struct _dircontents *) NULL)
	return (struct dirent *) NULL;

    dp.d_reclen = strlen (strcpy (dp.d_name, dirp->dd_cp->_d_entry));
    dp.d_off    = dirp->dd_loc * 32;
    dp.d_ino    = (ino_t)++dirp->dd_loc;
    dirp->dd_cp = dirp->dd_cp->_d_next;
    strlwr (dp.d_name);

    return &dp;
}

void	rewinddir (dirp)
DIR	*dirp;
{
    seekdir (dirp, (off_t)0);
}

void	seekdir (dirp, off)
DIR	*dirp;
off_t	off;
{
    long		i = off;
    struct _dircontents	*dp;

    if (off < 0L)
	return;

    for (dp = dirp->dd_contents ; --i >= 0 && dp ; dp = dp->_d_next)
	;

    dirp->dd_loc = off - (i + 1);
    dirp->dd_cp = dp;
}

off_t	telldir(dirp)
DIR	*dirp;
{
    return dirp->dd_loc;
}

static void		free_dircontents(dp)
struct _dircontents	*dp;
{
    struct _dircontents	*odp;

    while (dp) 
    {
	if (dp->_d_entry)
	    free(dp->_d_entry);

	dp = (odp = dp)->_d_next;
	free((char *) odp);
    }
}
