/* MS-DOS SHELL - Parse Tree Executor
 *
 * MS-DOS SHELL - Copyright (c) 1990 Data Logic Limited and Charles Forsyth
 *
 * This code is based on (in part) the shell program written by Charles
 * Forsyth and is subject to the following copyright restrictions:
 *
 * 1.  Redistribution and use in source and binary forms are permitted
 *     provided that the above copyright notice is duplicated in the
 *     source form and the copyright notice in file sh6.c is displayed
 *     on entry to the program.
 *
 * 2.  The sources (or parts thereof) or objects generated from the sources
 *     (or parts of sources) cannot be sold under any circumstances.
 *
 *    $Header: sh3.c 1.1 90/01/25 13:41:24 MS_user Exp $
 *
 *    $Log:	sh3.c $
 * Revision 1.1  90/01/25  13:41:24  MS_user
 * Initial revision
 * 
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <process.h>
#include <dos.h>
#include <signal.h>
#include <errno.h>
#include <setjmp.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <limits.h>

#include "sh.h"

/* static Function and string declarations */

static int	forkexec (C_Op *, int, int, int, char **);
static bool	iosetup (IO_Actions *, int, int);
static C_Op	**find1case (C_Op *, char *);
static C_Op	*findcase (C_Op *, char *);
static void	echo (char **);
static void	setsig (int, int (*)());
static int	rexecve (char *, char **, char **, bool);
static int	Execute_program (char *, char **, char **, bool);
static int	S_spawnve (char *, char **, char **);
static void	get_sys_info (void);
static void	EMS_error (char *, int);
static int	EMS_Close (void);
static int	build_command_line (char *, char **, char **);
static void	Clear_Extended_File (void);
static int	setstatus (int);

static char	*AE2big = "arg/env list too big";
static char	*EMS_emsg = "Warning: EMS Error (%x)\n";
			/* Extended Command line processing file name	*/
static char		*Extend_file = (char *)NULL;
static unsigned int	SW_EMsize;	/* Number of extend memory blks	*/

/*
 * execute tree recursively
 */

int		execute (t, pin, pout, act)
register C_Op	*t;
int		pin;
int		pout;
int		act;
{
    register C_Op	*t1;
    int			i, localpipe;
    char		*cp, **wp;
    char		**Local_Tword;
    Var_List		*vp;
    Break_C		bc;
    Break_C		*S_RList;		/* Save link pointers	*/
    Break_C		*S_BList;
    Break_C		*S_SList;
    int			Local_depth;		/* Save local values	*/
    int			Local_areanum;
    int			rv = 0;

/* End of tree ? */

    if (t == (C_Op *)NULL)
	return 0;

/* Save original and Increment execute function recursive level */

    Local_depth = Execute_stack_depth++;

/* Save original and increment area number */

    Local_areanum = areanum++;

/* Save the exit points from SubShells, functions and for/whiles */

    S_RList = Return_List;
    S_BList = Break_List;
    S_SList = SShell_List;

/* Expand any arguments */

    wp = (Local_Tword = t->words) != (char **)NULL
	 ? eval (Local_Tword, (t->type == TCOM) ? DOALL : DOALL & ~DOKEY)
	 : (char **)NULL;

/* Switch on tree node type */

    switch (t->type)
    {
	case TFUNC:			/* name () { list; }	*/
	    Save_Function (t, FALSE);
	    break;

/* In the case of a () command string, we need to save and restore the
 * current environment, directory and traps (can't think of anything else).
 * For any other, we just restore the current directory.  Also, we don't
 * want changes in the Variable list header saved for SubShells, because
 * we are effectively back at execute depth zero.
 */
	case TPAREN:			/* ()			*/
	    if ((rv = Create_NG_VL ()) == -1)
		break;

	    if (setjmp (bc.brkpt) == 0)
	    {
		Return_List = (Break_C *)NULL;
		Break_List  = (Break_C *)NULL;
		bc.nextlev  = SShell_List;
		SShell_List = &bc;
		rv = forkexec (t, pin, pout, act, wp);
	    }

/* Restore the original environment */

	    Return_List	= S_RList;
	    Break_List	= S_BList;
	    SShell_List	= S_SList;
	    Restore_Environment (rv, Local_depth);
	    break;

/* After a normal command, we need to restore the original directory.  Note
 * that a cd will have updated the variable $~, so no problem
 */

	case TCOM:			/* A command process	*/
	    rv = forkexec (t, pin, pout, act, wp);
	    Restore_Dir ();
	    break;

	case TPIPE:			/* Pipe processing		*/
	    if ((rv = openpipe ()) < 0)
		break;

/* Create pipe, execute command, reset pipe, execute the other side, close
 * the pipe and fini
 */

	    localpipe = remap (rv);
	    execute (t->left, pin, localpipe, 0);
	    lseek (localpipe, 0L, SEEK_SET);
	    rv = execute (t->right, localpipe, pout, 0);
	    closepipe (localpipe);
	    break;

	case TLIST:			/* Entries in a for statement	*/
	    execute (t->left, pin, pout, 0);
	    rv = execute (t->right, pin, pout, 0);
	    break;

	case TASYNC:			/* Async - not supported	*/
	    rv = -1;
	    S_puts ("sh: Async commands not supported\n");
	    setstatus (rv);
	    break;

	case TOR:			/* || and &&			*/
	case TAND:
	    rv = execute (t->left, pin, pout, 0);

	    if (((t1 = t->right) != (C_Op *)NULL) &&
		((rv == 0) == (t->type == TAND)))
		rv = execute (t1, pin, pout, 0);

	    break;

	case TFOR:			/* First part of a for statement*/

/* for x do...done - use the parameter values.  Need to know how many as
 * it is not a NULL terminated array
 */

	    if (wp == (char **)NULL)
	    {
		wp = dolv + 1;

		if ((i = dolc) < 0)
		    i = 0;
	    }

/* for x in y do...done - find the start of the variables and use them all */

	    else
	    {
		i = -1;
		while (*wp++ != (char *)NULL)
		    ;
	    }

/* Create the loop variable. */

	    vp = lookup (t->str, TRUE);

/* Set up a long jump return point before executing the for function so that
 * the continue statement is executed, ie we reprocessor the for condition.
 */

	    while (rv = setjmp (bc.brkpt))
	    {

/* Restore the current stack level and clear out any I/O */

		Restore_Environment (0, Local_depth + 1);
		Return_List = S_RList;
		SShell_List = S_SList;

/* If this is a break - clear the variable and terminate the while loop and
 * switch statement
 */

		if (rv == BC_BREAK)
		    break;
	    }

	    if (rv == BC_BREAK)
		break;

/* Process the next entry - Add to the break/continue chain */

	    bc.nextlev = Break_List;
	    Break_List = &bc;

/* Execute the command tree */

	    for (t1 = t->left; i-- && *wp != NULL;)
	    {
		setval (vp, *wp++);
		rv = execute (t1, pin, pout, 0);
	    }

/* Remove this tree from the break list */

	    Break_List = S_BList;
	    break;

/* While and Until function.  Similar to the For function.  Set up a
 * long jump return point before executing the while function so that
 * the continue statement is executed OK.
 */

	case TWHILE:			/* WHILE and UNTIL functions	*/
	case TUNTIL:
	    while (rv = setjmp (bc.brkpt))
	    {

/* Restore the current stack level and clear out any I/O */

		Restore_Environment (0, Local_depth + 1);
		Return_List = S_RList;
		SShell_List = S_SList;

/* If this is a break, terminate the while and switch statements */

		if (rv == BC_BREAK)
		    break;
	    }

	    if (rv == BC_BREAK)
		break;

/* Set up links */

	    bc.nextlev = Break_List;
	    Break_List = &bc;
	    t1 = t->left;

	    while ((execute(t1, pin, pout, 0) == 0) == (t->type == TWHILE))
		rv = execute (t->right, pin, pout, 0);

	    Break_List = S_BList;
	    break;

	case TIF:			/* IF and ELSE IF functions	*/
	case TELIF:
	    rv = !execute (t->left, pin, pout, 0)
	   		 ? execute (t->right->left, pin, pout, 0)
			 : execute (t->right->right, pin, pout, 0);
	    break;

	case TCASE:			/* CASE function		*/
	    if ((cp = evalstr (t->str, DOSUB | DOTRIM)) == (char *)NULL)
		cp = null;

	    if ((t1 = findcase (t->left, cp)) != (C_Op *)NULL)
		rv = execute (t1, pin, pout, 0);

	    break;

	case TBRACE:			/* {} statement			*/
	    if ((rv >= 0) && ((t1 = t->left) != (C_Op *)NULL))
		rv = execute (t1, pin, pout, 0);

	    break;
    }

/* Processing Completed - Restore environment */

    t->words		= Local_Tword;
    Execute_stack_depth = Local_depth;

/* Remove unwanted malloced space */

    freehere (areanum);
    freearea (areanum);

    areanum		= Local_areanum;

/* Check for interrupts */

    if (talking && SW_intr)
    {
	closeall ();
	fail ();
    }

/* Check for traps */

    if ((i = trapset) != 0)
    {
	trapset = 0;
	runtrap (i);
    }

    return rv;
}

/*
 * Restore the original directory
 */

void	Restore_Dir ()
{
    unsigned int	dummy;

    _dos_setdrive (tolower(*C_dir->value) - 'a' + 1, &dummy);

    if (chdir (&C_dir->value[2]) != 0)
    {
	S_puts ("Warning: current directory reset to /\n");
	chdir ("/");
	Getcwd ();
    }
}

/*
 * Ok - execute the program, resetting any I/O required
 */

static int	forkexec (t, pin, pout, act, wp)
register C_Op	*t;
int		pin;
int		pout;
int		act;
char		**wp;
{
    int		rv = -1;
    int		(*shcom)(C_Op *) = (int (*)())NULL;
    char	*cp;
    IO_Actions	**iopp;
    int		resetsig = 0;
    char	**owp = wp;
    bool	spawn = FALSE;
    Fun_Ops	*fop;

    if (t->type == TCOM)
    {
	while ((cp = *wp++) != (char *)NULL)
	    ;

	cp = *wp;

/* strip all initial assignments not correct wrt PATH=yyy command  etc */

	if (FL_TEST ('x'))
	    echo (cp != (char *)NULL ? wp : owp);

/* Is it only an assignement? */

	if ((cp == (char *)NULL) && (t->ioact == (IO_Actions **)NULL))
	{
	    while (((cp = *owp++) != (char *)NULL) && assign (cp, COPYV))
		;

	    return setstatus (0);
	}

/* Check for built in commands */

	else if (cp != (char *)NULL)
	    shcom = inbuilt (cp);
    }

/* Unix fork simulation? */

    t->words = wp;
    if (shcom == NULL && (act & FEXEC) == 0)
    {
	spawn = TRUE;

	if (talking)
	{
#ifdef SIGQUIT
	    signal (SIGQUIT, SIG_IGN);
#endif
	    signal (SIGINT, SIG_IGN);
	    resetsig = 1;
	}
    }

/* Set any variables */

    while (((cp = *owp++) != (char *)NULL) && assign (cp, COPYV))
    {
	if (shcom == NULL)
	    s_vstatus (lookup (cp, TRUE), EXPORT);
    }

/* We cannot close the pipe, because once the exec/spawn has taken place
 * the processing of the pipe is not yet complete.
 */

    if (pin != NOPIPE)
    {
	S_dup2 (pin, STDIN_FILENO);
	lseek (STDIN_FILENO, 0L, SEEK_SET);
    }

    if (pout != NOPIPE)
    {
	S_dup2 (pout, STDOUT_FILENO);
	lseek (STDOUT_FILENO, 0L, SEEK_END);
    }

/* Set up any other IO required */

    if ((iopp = t->ioact) != (IO_Actions **)NULL)
    {
	while (*iopp != (IO_Actions *)NULL)
	{
	    if (iosetup (*iopp++, pin, pout))
		return rv;
	}
    }

    if (shcom)
	return restore_std (setstatus ((*shcom)(t)));

/* All fids above 10 are autoclosed in the exec file because we have used
 * the O_NOINHERIT flag.  Note I patched open.obj to pass this flag to the
 * open function.
 */

    if (resetsig)
    {
#ifdef SIGQUIT
	signal (SIGQUIT, SIG_IGN);
#endif
	signal (SIGINT, onintr);
    }

    if (t->type == TPAREN)
	return restore_std (execute (t->left, NOPIPE, NOPIPE, FEXEC));

/* Are we just changing the I/O re-direction for the shell ? */

    if (wp[0] == NULL)
    {
	if (spawn)
	    restore_std (0);

	return 0;
    }

/* No - Check for a function the program.  At this point, we need to put
 * in some processing for return.
 */

    if ((fop = Fun_Search (wp[0])) != (Fun_Ops *)NULL)
    {
	char			**s_dolv = dolv;
	int			s_dolc   = dolc;
	Break_C			*s_RList = Return_List;
	Break_C			*s_BList = Break_List;
	Break_C			*s_SList = SShell_List;
	int			LS_depth = Execute_stack_depth;
	Break_C			bc;

/* Set up $0..$n for the function */

	dolv = wp;
	for (dolc = 0; dolv[dolc] != (char *)NULL; ++dolc);
	setval (lookup ("#", TRUE), putn (dolc));

	if (setjmp (bc.brkpt) == 0)
	{
	    Break_List  = (Break_C *)NULL;
	    bc.nextlev  = Return_List;
	    Return_List = &bc;
	    rv = execute (fop->tree->left, NOPIPE, NOPIPE, FEXEC);
	}

/* A return has been executed - Unlike, while and for, we just need to
 * restore the local execute stack level and the return will restore
 * the correct I/O.
 */

	else
	    rv = getn (lookup ("?", FALSE)->value);

/* Restore the old $0, and previous return address */

	Break_List  = s_BList;
	Return_List = s_RList;
	SShell_List = s_SList;
	dolv	    = s_dolv;
	dolc	    = s_dolc;
	Restore_Environment (rv, LS_depth);
	setval (lookup ("#", TRUE), putn (dolc));
	return rv;
    }

/* Ok - execute the program */

    return restore_std (rexecve (wp[0], wp, makenv (), spawn));
}

/*
 * Restore Local Environment
 */

void	Restore_Environment (retval, stack)
int	retval;
int	stack;
{
    Execute_stack_depth = stack;
    Delete_G_VL ();
    Restore_Dir ();
    restore_std (setstatus (retval));
}

/*
 * Set up I/O redirection.  0< 1> are ignored as required within pipelines.
 */

static bool		iosetup (iop, pipein, pipeout)
register IO_Actions	*iop;
int			pipein;
int			pipeout;
{
    register int	u;
    char		*cp, *msg;

    if (iop->io_unit == IODEFAULT)	/* take default */
	iop->io_unit = (iop->io_flag & (IOREAD | IOHERE)) ? STDIN_FILENO
							  : STDOUT_FILENO;

/* Check for pipes */

    if ((pipein != NOPIPE) && (iop->io_unit == STDIN_FILENO))
	return FALSE;

    if ((pipeout != NOPIPE) && (iop->io_unit == STDOUT_FILENO))
	return FALSE;

    msg = (iop->io_flag & (IOREAD | IOHERE)) ? "open" : "create";

    if ((iop->io_flag & IOHERE) == 0)
    {
	if ((cp = evalstr (iop->io_name, DOSUB | DOTRIM)) == (char *)NULL)
	    return TRUE;
    }

    if (iop->io_flag & IODUP)
    {
	if ((cp[1]) || !isdigit (*cp) && *cp != '-')
	{
	    print_error ("%s: illegal >& argument\n", cp);
	    return TRUE;
	}

	if (*cp == '-')
	    iop->io_flag = IOCLOSE;

	iop->io_flag &= ~(IOREAD | IOWRITE);
    }

/* Open the file in the appropriate mode */

    switch (iop->io_flag)
    {
	case IOREAD:				/* <			*/
	    u = S_open (FALSE, cp, O_RDONLY);
	    break;

	case IOHERE:				/* <<			*/
	case IOHERE | IOXHERE:
	    u = herein (iop->io_name, iop->io_flag & IOXHERE);
	    cp = "here file";
	    break;

	case IOWRITE | IOCAT:			/* >>			*/
	    if (check_rsh (cp))
		return TRUE;

	    if ((u = S_open (FALSE, cp, O_WRONLY | O_TEXT)) >= 0)
	    {
		lseek (u, 0L, SEEK_END);
		break;
	    }

	case IOWRITE:				/* >			*/
	    if (check_rsh (cp))
		return TRUE;

	    u = S_open (FALSE, cp, O_CMASK, 0666);
	    break;

	case IODUP:				/* >&			*/
	    if (check_rsh (cp))
		return TRUE;

	    u = S_dup2 (*cp - '0', iop->io_unit);
	    break;

	case IOCLOSE:				/* >-			*/
	    if ((iop->io_unit >= STDIN_FILENO) &&
		(iop->io_unit <= STDERR_FILENO))
		S_dup2 (-1, iop->io_unit);

	    S_close (iop->io_unit, TRUE);
	    return FALSE;
    }

    if (u < 0)
    {
	print_warn ("%s: cannot %s\n", cp, msg);
	return TRUE;
    }

    else if (u != iop->io_unit)
    {
	S_dup2 (u, iop->io_unit);
	S_close (u, TRUE);
    }

    return FALSE;
}

/*
 * -x flag - echo command to be executed
 */

static void	echo (wp)
register char	**wp;
{
    register int	i;

    S_putc ('+');

    for (i = 0; wp[i] != (char *)NULL; i++)
    {
	S_putc (SP);
	S_puts (wp[i]);
    }

    S_putc (NL);
}

static C_Op	**find1case (t, w)
C_Op		*t;
char		*w;
{
    register C_Op	*t1;
    C_Op		**tp;
    register char	**wp, *cp;

    if (t == (C_Op *)NULL)
	return (C_Op **)NULL;

    if (t->type == TLIST)
    {
	if ((tp = find1case (t->left, w)) != (C_Op *)NULL)
	    return tp;

	t1 = t->right;	/* TPAT */
    }

    else
	t1 = t;

    for (wp = t1->words; *wp != (char *)NULL;)
    {
	if ((cp = evalstr (*(wp++), DOSUB)) && gmatch (w, cp, FALSE))
	    return &t1->left;
    }

    return (C_Op **)NULL;
}

static C_Op	*findcase (t, w)
C_Op		*t;
char		*w;
{
    register C_Op **tp;

    return ((tp = find1case (t, w)) != (C_Op **)NULL) ? *tp : (C_Op *)NULL;
}

/*
 * Set up the status on exit from a command
 */

static int	setstatus (s)
register int	s;
{
    exstat = s;
    setval (lookup ("?", TRUE), putn (s));
    return s;
}

/*
 * PATH-searching interface to execve.  If getenv ("PATH") were kept
 * up-to-date, execvp might be used.
 */

static int	rexecve (c, v, envp, d_flag)
char		*c;
char		**v;
char		**envp;
bool		d_flag;
{
    register char	*sp;
    int			res;
    char		*em;
    bool		eloop;

/* If the environment is null - It is too big - error */

    if (envp == (char **)NULL)
	em = AE2big;

    else
    {
	sp = any ('/', c) ? null : path->value;

	do
	{
	    sp = path_append (sp, c, e.linep);

	    if ((res = Execute_program (e.linep, v, envp, d_flag)) != -1)
		return res;

	    eloop = TRUE;

	    switch (errno)
	    {

/* No entry for the file - if the file exists, execute it as a shell
 * script
 */
		case ENOENT:
		    if ((res = O_for_execute (e.linep)) >= 0)
		    {
			S_close (res, TRUE);
			*v = e.linep;
			em = *--v;
			*v = e.linep;
		        res = Execute_program (lookup (shell, FALSE)->value,
					       v, envp, d_flag);
			*v = em;

			if (res != -1)
			return res;

			em = "no Shell";
		    }

		    else
			em = "not found";

		    eloop = FALSE;
		    break;

		case ENOEXEC:
		    em = "program corrupt";
		    break;

		case ENOMEM:
		    em = "program too big";
		    break;

		case E2BIG:
		    em = AE2big;
		    break;

		default:
		    em = "cannot execute";
		    eloop = FALSE;
		    break;
	    }
	} while ((sp != (char *)NULL) && !eloop);
    }

    print_warn ("%s: %s\n", c, em);

    if (!d_flag)
	exit (-1);

    return -1;
}

/*
 * Run the command produced by generator `f' applied to stream `arg'.
 */

int		run (argp, f)
IO_Args		*argp;
int		(*f)(IO_State *);
{
    Word_B		*swdlist = wdlist;
    Word_B		*siolist = iolist;
    jmp_buf		ev, rt;
    int			*ofail = failpt;
    int			rv = -1;
    Break_C		*S_RList = Return_List;	/* Save loval links	*/
    Break_C		*S_BList = Break_List;
    Break_C		*S_SList = SShell_List;
    Break_C		bc;
    int			LS_depth = Execute_stack_depth;
    C_Op		*outtree;

/* Create a new environment in which to run */

    if (Create_NG_VL () == -1)
	return -1;

/* Create a new save area */

    areanum++;

/* Execute the command */

    if (newenv (setjmp (errpt = ev)) == FALSE)
    {
	Return_List = (Break_C *)NULL;
	Break_List  = (Break_C *)NULL;
	wdlist      = (Word_B *)NULL;
	iolist      = (Word_B *)NULL;

	pushio (argp, f);
	e.iobase = e.iop;
	yynerrs = 0;


	if ((setjmp (failpt = rt) == 0) &&
	    ((outtree = yyparse ()) != (C_Op *)NULL))
	{
	    if (setjmp (bc.brkpt) == 0)
	    {
		bc.nextlev = SShell_List;
		SShell_List = &bc;
		rv = execute (outtree, NOPIPE, NOPIPE, 0);
	    }

	    else
		rv = getn (lookup ("?", FALSE)->value);
	}

	quitenv ();
    }

/* Restore the environment */

    Return_List = S_RList;
    Break_List = S_BList;
    SShell_List = S_SList;
    wdlist = swdlist;
    iolist = siolist;
    failpt = ofail;

    Restore_Environment (rv, LS_depth);

    freearea (areanum--);
    return rv;
}

/* Exec or spawn the program ? */

static int	Execute_program (path, parms, envp, d_flag)
char		*path;
char		**parms;
char		**envp;
bool		d_flag;
{
    return setstatus ((!d_flag) ? execve (path, parms, envp)
				: S_spawnve (path, parms, envp));
}

/* Set up to spawn a process */

static int	S_spawnve (path, parms, envp)
char		*path;
char		**parms;
char		**envp;
{
    unsigned int	c_cur = (unsigned int)(_psp - 1);
    unsigned int	size = 0;
    char		*ep, *ep1;
    int			res, serrno;
    struct MCB_list	*mp = (struct MCB_list *)((unsigned long)c_cur << 16L);


/* Check to see if the file exists */

    strcpy (path_line, path);

    if ((ep = strrchr (path_line, '/')) == (char *)NULL)
	ep = path_line;

/* If no dot in name - check for .exe and .com files */

    if ((ep1 = strchr (ep, '.')) == (char *)NULL)
    {
	ep1 = ep + strlen (ep);
	strcpy (ep1, ".exe");

	if ((res = access (path_line, F_OK)) != 0)
	{
	    strcpy (ep1, ".com");
	    res = access (path_line, F_OK);
	}

	if (res != 0)
	    return -1;
    }

    else if ((stricmp (ep1, ".exe") != 0) && (stricmp (ep1, ".com") != 0))
    {
	errno = ENOEXEC;
	return -1;
    }

    else if (access (path_line, F_OK) != 0)
	return -1;

/* Process the command line.  If no swapping, we have executed the program */

    res = build_command_line (path_line, parms, envp);

    if ((Swap_Mode == SWAP_OFF) || res)
	return res;

/* Find the length of the swap area */

    while ((mp = (struct MCB_list *)((unsigned long)c_cur << 16L))->MCB_type
	    == MCB_CON)
    {
	if ((mp->MCB_pid != _psp) && (mp->MCB_pid != 0) &&
	    (mp->MCB_type != MCB_END))
	{
	    Clear_Extended_File ();
	    print_error ("Fatal: Memory chain corrupt\n");
	    return -1;
	}

	c_cur += (mp->MCB_len + 1);
	size += mp->MCB_len + 1;
    }

/*
 * Convert swap size from paragraphs to 16K blocks.
 */

    if (size == 0)
	size = mp->MCB_len + 1;

    SW_Blocks = (size / 0x0400) + 1;

/* OK Now we've set up the FCB's, command line and opened the swap file.
 * Get some sys info for the swapper and execute my little assembler
 * function to swap us out
 */

    get_sys_info ();

/* Ok - 3 methods of swapping */

/* If expanded memory - try that */

    if (Swap_Mode & SWAP_EXPAND)
    {
	int	cr;
	SW_Mode = 3;			/* Set Expanded memory swap	*/

	res = SA_spawn (envp);
	cr = EMS_Close ();		/* Close EMS			*/

	if ((res != -2) && cr)		/* Report Close error ?		*/
	{
	    res = -2;
	    errno = cr;
	}

	if (res == -2)
	    EMS_error ("Expanded memory swap failed (%x)\n", errno);

	else
	{
	    Clear_Extended_File ();
	    return res;
	}

/* Failed - disabled */

	Swap_Mode &= (~SWAP_EXPAND);
    }

    if (Swap_Mode & SWAP_EXTEND)
    {
	SW_Mode = 2;			/* Set Extended memory swap	*/

        if ((SW_EMsize <= SW_Blocks) ||
	    ((SW_EMstart - 0x100000L +
	      ((long)(SW_Blocks - SW_EMsize) * 16L * 1024L)) < 0L))
	    print_warn ("Not enough Extended memory for swap\n");

	else if ((res = SA_spawn (envp)) == -2)
	    print_warn ("Extended memory swap failed (%x)\n", errno);

	else
	{
	    Clear_Extended_File ();
	    return res;
	}

/* Failed - disabled */

	Swap_Mode &= (~SWAP_EXTEND);
    }

/* Try the disk if available */

    if (Swap_Mode & SWAP_DISK)
    {
	if ((SW_fp = S_open (TRUE, g_tempname (), O_SMASK, 0600)) < 0)
	{
	    print_error ("No Swap files\n");
	    errno = ENOSPC;
	    return -1;
	}

	SW_Mode = 1;			/* Set Disk file swap		*/

/* Execute the program */

	res = SA_spawn (envp);

	Clear_Extended_File ();

	if (res == -2)
	{
	    print_warn ("Swap file write failed\n");
	    errno = ENOSPC;
	    res = -1;
	}

/* Close the swap file and return the result */

	serrno = errno;
	S_close (SW_fp, TRUE);
	errno = serrno;
	return res;
    }

/* No swapping available - give up */

    Clear_Extended_File ();
    print_error ("All Swapping methods failed\n");
    errno = ENOSPC;
    return -1;
}

/* Get some system info */

static void	get_sys_info ()
{
    union REGS		or;
    struct SREGS	sr;
    char		*sp;

/* Save the interrupt 0 address */

    or.x.ax = 0x3500;
    intdosx (&or, &or, &sr);

    SW_I0_V_BX = or.x.bx;
    SW_I0_V_ES = sr.es;

/* Save the interrupt 23 address */

    or.x.ax = 0x3523;
    intdosx (&or, &or, &sr);

    SW_I23_V_BX = or.x.bx;
    SW_I23_V_ES = sr.es;

/* Get max Extended memory pages, and convert to 16K blocks.  If Extended
 * memory swapping disabled, set to zero
 */

    or.x.ax = 0x8800;
    int86 (0x15, &or, &or);
    SW_EMsize = (Swap_Mode & SWAP_EXTEND) ? or.x.ax / 16 : 0;

/* Check for the Expand Memory System */

    if (!(Swap_Mode & SWAP_EXPAND))
	return;

    SW_fp = -1;				/* Set EMS handler not defined	*/

    or.x.ax = 0x3567;
    intdosx (&or, &or, &sr);

    sp = (char *)((unsigned long)(sr.es) << 16L | 10L);

/* If not there - disable */

    if (memcmp ("EMMXXXX0", sp, 8) != 0)
    {
	EMS_error ("Warning: EMS not available\n", 0);
	return;
    }

    or.h.ah = 0x40;			/* Check status			*/
    int86 (0x67, &or, &or);

    if (or.h.ah != 0)
    {
	EMS_error (EMS_emsg, or.h.ah);
	return;
    }

/* Check version greater than 3.2 */

    or.h.ah = 0x46;
    int86 (0x67, &or, &or);

    if ((or.h.ah != 0) || (or.h.al < 0x32))
    {
	EMS_error (EMS_emsg, or.h.ah);
	return;
    }

/*  get page frame address */

    or.h.ah = 0x41;
    int86 (0x67, &or, &or);

    if (or.h.ah != 0)
    {
	EMS_error (EMS_emsg, or.h.ah);
	return;
    }

    SW_EMSFrame = or.x.bx;		/* Save the page frame		*/

/* Get the number of pages required */

    or.h.ah = 0x43;
    or.x.bx = SW_Blocks;
    int86 (0x67, &or, &or);

    if (or.h.ah != 0)
    {
	EMS_error (EMS_emsg, or.h.ah);
	return;
    }

/* Save the EMS Handler */

    SW_fp = or.x.dx;

/* save EMS page map */

    or.h.ah = 0x47;
    or.x.dx = SW_fp;
    int86 (0x67, &or, &or);

    if (or.h.ah != 0)
    {
	EMS_error (EMS_emsg, or.h.ah);
	return;
    }
}

/* Print EMS error message */

static void	EMS_error (s, v)
char		*s;
int		v;
{
    print_warn (s, v);
    Swap_Mode &= ~(SWAP_EXPAND);
    EMS_Close ();
}


/* If the handler is defined - close it */

static int	EMS_Close ()
{
    union REGS		or;
    int			res = 0;

    if (SW_fp == -1)
	return 0;

/* Restore EMS page */

    or.h.ah = 0x48;
    or.x.dx = SW_fp;
    int86 (0x67, &or, &or);

    if (or.h.ah != 0)
	res = or.h.al;

    or.h.ah = 0x45;
    or.x.dx = SW_fp;
    int86 (0x67, &or, &or);

    SW_fp = -1;
    return (res) ? res : or.h.ah;
}

/* Set up command line.  If the EXTENDED_LINE variable is set, we create
 * a temporary file, write the argument list (one entry per line) to the
 * this file and set the command line to @<filename>.  If NOSWAPPING, we
 * execute the program because I have to modify the argument line
 */

int	build_command_line (path, argv, envp)
char	*path;
char	**argv;
char	**envp;
{
    char		**pl = argv;
    char		*fname;
    int			res, fd;
    char		*pname;
    FILE		*fp;
    char		nbuffer[NAME_MAX + 2];
    bool		found;
    char		*ep;
    char		*new_args[3];

/* Find the start of the program name */

    if ((pname = strrchr (path, '/')) == (char *)NULL)
	pname = path;

    else
	++pname;

/* Translate process name to MSDOS format */

    Convert_Slashes (path);
    strupr (path);

/* Extended command line processing */

    Extend_file == (char *)NULL;		/* Set no file		*/

    if ((*(pl++) != (char *)NULL) &&
	((fname = lookup ("EXTENDED_LINE", FALSE)->value) != null) &&
	((fp = fopen (fname, "rt")) != (FILE *)NULL))
    {

/* Loop through the file look for the current program */

	found = FALSE;

	while (fgets (nbuffer, NAME_MAX + 1, fp) != (char *)NULL)
	{
	    if ((ep = strchr (nbuffer, '\n')) != (char *)NULL)
		*ep = 0;

	    if (stricmp (nbuffer, pname) == 0)
	    {
		found = TRUE;
		break;
	    }
	}

	fclose (fp);

/* Check parameters don't contain a re-direction parameter */

	if (found)
	{
	    char	**pl1 = pl;

	    while (*pl1 != (char *)NULL)
	    {
		if (**(pl1++) == '@')
		{
		    found = FALSE;
		    break;
		}
	    }
	}

/* If we find it - create a temporary file and write the stuff */

	if ((found) &&
	    ((fd = S_open (FALSE, Extend_file = g_tempname (), O_CMASK,
			   0600)) >= 0))
	{

/* Copy to end of list */

	    while (*pl != (char *)NULL)
	    {
		if (((res = strlen (*pl)) && (write (fd, *pl, res) != res)) ||
		    (write (fd, "\n", 1) != 1))
		{
		    close (fd);
		    unlink (Extend_file);
		    Extend_file == (char *)NULL;
		    errno = ENOSPC;
		    return -1;
		}

		++pl;
	    }

/* Completed write OK */

	    close (fd);

/* Set up cmd_line[1] to contain the filename */

	    memset (cmd_line, 0, CMD_LINE_MAX);
	    cmd_line[1] = '@';
	    strcpy (&cmd_line[2], Extend_file);
	    cmd_line[0] = (char)(strlen (Extend_file) + 1);

/* Correctly terminate cmd_line in no swap mode */

	    if (Swap_Mode != SWAP_OFF)
		cmd_line[cmd_line[0] + 1] = '\r';

/* If the name in the file is in upper case - use \ for separators */

	    if (isupper (*nbuffer))
		Convert_Slashes (&cmd_line[2]);

/* OK we are ready to execute */

	    if (Swap_Mode == SWAP_OFF)
	    {
		new_args[0] = *argv;
		new_args[1] = &cmd_line[1];
		new_args[2] = (char *)NULL;
		return spawnve (P_WAIT, path, new_args, envp);
	    }

	    else
		return 0;
	}
    }

/* Check length of Parameter list */

    res = 0;
    cmd_line[0] = 0;
    cmd_line[1] = '\r';
    ep = cmd_line;

/* Skip the first parameter and get the length of the rest */

    if (*argv != (char *)NULL)
    {
	while (*pl != (char *)NULL)
	{
	    if ((res += (strlen (*pl) + 1)) >= CMD_LINE_MAX)
	    {
		errno = E2BIG;
		return -1;
	    }

	    strcat (strcat (ep, " "), *(pl++));
	}

	if (res)
	    cmd_line[res--] = '\r';
    }

/* Terminate the line and insert the line length */

    cmd_line[0] = (char)res;

/* If swapping disabled - just execute it */

    return (Swap_Mode == SWAP_OFF) ? spawnve (P_WAIT, path, argv, envp) : 0;
}

/* Clear Extended command line file */

static void	Clear_Extended_File ()
{
    if (Extend_file != (char *)NULL)
	unlink (Extend_file);

    Extend_file = (char *)NULL;
}
