/* MS-DOS SHELL - Unix File I/O Emulation
 *
 * MS-DOS SHELL - Copyright (c) 1990 Data Logic Limited
 *
 * This code is subject to the following copyright restrictions:
 *
 * 1.  Redistribution and use in source and binary forms are permitted
 *     provided that the above copyright notice is duplicated in the
 *     source form and the copyright notice in file sh6.c is displayed
 *     on entry to the program.
 *
 * 2.  The sources (or parts thereof) or objects generated from the sources
 *     (or parts of sources) cannot be sold under any circumstances.
 *
 *    $Header: sh8.c 1.1 90/01/29 17:46:37 MS_user Exp $
 *
 *    $Log:	sh8.c $
 * Revision 1.1  90/01/29  17:46:37  MS_user
 * Initial revision
 * 
 * 
 */

#include <sys/types.h>
#include <signal.h>
#include <errno.h>
#include <setjmp.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include "sh.h"

#define F_START		4

static char		*nopipe = "can't create pipe - try again\n";

/* List of open files to allow us to simulate the Unix open and unlink
 * operation for temporary files
 */

typedef struct flist {
    struct flist	*fl_next;	/* Next link			*/
    char		*fl_name;	/* File name			*/
    bool		fl_close;	/* Delete on close flag		*/
    int			fl_size;	/* Size of fl_fd array		*/
    int			fl_count;	/* Number of entries in array	*/
    int			fl_mode;	/* File open mode		*/
    int			*fl_fd;		/* File ID array (for dup)	*/
} s_flist;

static s_flist		*list_start = (s_flist *)NULL;
static s_flist		*find_entry (int);

/*
 * Open a file and add it to the Open file list.  Errors are the same as
 * for a normal open.
 */

int	S_open (d_flag, name, mode)
bool	d_flag;
char	*name;
int	mode;
{
    va_list		ap;
    int			pmask;
    s_flist		*fp = (struct s_flist *)NULL;
    int			*f_list = (int *)NULL;
    char		*f_name = (char *)NULL;

/* Check the permission mask if it exists */

    va_start (ap, mode);
    pmask =  va_arg (ap, int);
    va_end (ap);

/* Grap some space.  If it fails, free space and return an error */

    if (((fp = (s_flist *) space (sizeof (s_flist))) == (s_flist *)NULL) ||
	((f_list = (int *) space (sizeof (int) * F_START)) == (int *)NULL) ||
	((f_name = strsave (name, 0)) == null))
    {
	if (f_list == (int *)NULL)
	    DELETE (f_list);

	if (fp == (s_flist *)NULL)
	    DELETE (fp);

	errno = ENOMEM;
	return -1;
    }

/* Set up the structure */

    fp->fl_name  = strcpy (f_name, name);
    fp->fl_close = d_flag;
    fp->fl_size  = F_START;
    fp->fl_count = 1;
    fp->fl_fd    = f_list;
    fp->fl_mode  = mode;

/* Open the file */

    if ((fp->fl_fd[0] = open (name, mode, pmask)) < 0)
    {
	pmask = errno;
	DELETE (f_name);
	DELETE (f_list);
	DELETE (fp);
	errno = pmask;
	return -1;
    }

/* Make sure everything is in area 0 */

    setarea ((char *)fp, 0);
    setarea ((char *)f_list, 0);

/* List into the list */

    fp->fl_next   = list_start;
    list_start = fp;

/* Return the file descriptor */

    return fp->fl_fd[0];
}

/*
 * Scan the File list for the appropriate entry for the specified ID
 */

static s_flist		*find_entry (fid)
int			fid;
{
    s_flist	*fp = list_start;
    int		i;

    while (fp != (s_flist *)NULL)
    {
	for (i = 0; i < fp->fl_count; i++)
	{
	    if (fp->fl_fd[i] == fid)
		return fp;
	}

	fp = fp->fl_next;
    }

    return (s_flist *)NULL;
}

/* Close the file
 *
 * We need a version of close that does everything but close for dup2 as
 * new file id is closed.  If c_flag is TRUE, close the file as well.
 */

int	S_close (fid, c_flag)
int	fid;
bool	c_flag;
{
    s_flist	*fp = find_entry (fid);
    s_flist	*last = (s_flist *)NULL;
    s_flist	*fp1 = list_start;
    int		i, serrno;
    bool	release = TRUE;
    bool	delete = FALSE;
    char	*fname;

/* Find the entry for this ID */

    if (fp != (s_flist *)NULL)
    {
	for (i = 0; i < fp->fl_count; i++)
	{
	    if (fp->fl_fd[i] == fid)
		fp->fl_fd[i] = -1;

	    if (fp->fl_fd[i] != -1)
		release = FALSE;
	}

/* Are all the Fids closed ? */

	if (release)
	{
	    fname = fp->fl_name;
	    delete = fp->fl_close;
	    DELETE (fp->fl_fd);

/* Scan the list and remove the entry */

	    while (fp1 != (s_flist *)NULL)
	    {
		if (fp1 != fp)
		{
		    last = fp1;
		    fp1 = fp1->fl_next;
		    continue;
		}

		if (last == (s_flist *)NULL)
		    list_start = fp->fl_next;

		else
		    last->fl_next = fp->fl_next;

		break;
	    }

/* OK - delete the area */

	    DELETE (fp);
	}
    }

/* Close the file anyway */

    if (c_flag)
    {
	i = close (fid);
	serrno = errno;
    }

/* Delete the file ? */

    if (delete)
    {
	unlink (fname);
	DELETE (fname);
    }

/* Restore results and error code */

    errno = serrno;
    return i;
}

/*
 * Duplicate file handler.  Add the new handler to the ID array for this
 * file.
 */

int	S_dup (old_fid)
int	old_fid;
{
    int		new_fid;

    if ((new_fid = dup (old_fid)) >= 0)
	S_Remap (old_fid, new_fid);

    return new_fid;
}

/*
 * Add the ID to the ID array for this file
 */

void	S_Remap (old_fid, new_fid)
int	old_fid, new_fid;
{
    s_flist	*fp = find_entry (old_fid);
    int		*flist;
    int		i;

    if (fp == (s_flist *)NULL)
	return;

/* Is there an empty slot ? */

    for (i = 0; i < fp->fl_count; i++)
    {
	if (fp->fl_fd[i] == -1)
	{
	    fp->fl_fd[i] = new_fid;
	    return;
	}
    }

/* Is there any room at the end ? No - grap somemore space and effect a
 * re-alloc.  What to do if the re-alloc fails - should really get here.
 * Safty check only??
 */

    if (fp->fl_count == fp->fl_size)
    {
	if ((flist = (int *) space ((fp->fl_size + F_START) * sizeof (int)))
	    == (int *)NULL)
	    return;

	memcpy ((char *)flist, (char *)fp->fl_fd, sizeof (int) * fp->fl_size);
	DELETE (fp->fl_fd);

	fp->fl_fd   = flist;
	fp->fl_size += F_START;
    }

    fp->fl_fd[fp->fl_count++] = new_fid;
}

/*
 * Set Delete on Close flag
 */

void	S_Delete (fid)
int	fid;
{
    s_flist	*fp = find_entry (fid);

    if (fp != (s_flist *)NULL)
	fp->fl_close = TRUE;
}

/*
 * Duplicate file handler onto specific handler
 */

int	S_dup2 (old_fid, new_fid)
int	old_fid;
int	new_fid;
{
    int		res = 0;
    int		i;
    Save_IO	*sp;

/* If duping onto stdin, stdout or stderr, Search the Save IO stack for an
 * entry matching us
 */

    if ((new_fid >= STDIN_FILENO) && (new_fid <= STDERR_FILENO))
    {
	for (sp = SSave_IO, i = 0; (i < NSave_IO_E) &&
				   (SSave_IO[i].depth < Execute_stack_depth);
	     i++);

/* If depth is greater the Execute_stack_depth - we should panic as this
 * should not happen.  However, for the moment, I'll ignore it
 */

/* If there an entry for this depth ? */

	if (i == NSave_IO_E)
	{

/* Do we need more space? */

	    if (NSave_IO_E == MSave_IO_E)
	    {
		sp = (Save_IO *)space ((MSave_IO_E + SSAVE_IO_SIZE) * sizeof (Save_IO));

/* Check for error */

		if (sp == (Save_IO *)NULL)
		{
		    errno = ENOMEM;
		    return -1;
		}

/* Save original data */

		if (MSave_IO_E != 0)
		{
		    memcpy (sp, SSave_IO, sizeof (Save_IO) * MSave_IO_E);
		    DELETE (SSave_IO);
		}

		setarea ((char *)sp, 1);
		SSave_IO = sp;
		MSave_IO_E += SSAVE_IO_SIZE;
	    }

/* Initialise the new entry */

	    sp = &SSave_IO[NSave_IO_E++];
	    sp->depth             = Execute_stack_depth;
	    sp->fp[STDIN_FILENO]  = -1;
	    sp->fp[STDOUT_FILENO] = -1;
	    sp->fp[STDERR_FILENO] = -1;
	}

	if (sp->fp[new_fid] == -1)
	    sp->fp[new_fid] = remap (new_fid);
    }

/* OK - Dup the descriptor */

    if ((old_fid != -1) && ((res = dup2 (old_fid, new_fid)) >= 0))
    {
	S_close (new_fid, FALSE);
	S_Remap (old_fid, new_fid);
    }

    return res;
}

/*
 * Restore the Stdin, Stdout and Stderr to original values
 */

int	restore_std (rv)
int	rv;
{
    int		j, i;
    Save_IO	*sp;

/* Start at the top and remove any entries above the current execute stack
 * depth
 */

    for (j = NSave_IO_E; j > 0; j--)
    {
       sp = &SSave_IO[j - 1];

       if (sp->depth < Execute_stack_depth)
	   break;

/* Reduce number of entries */

	--NSave_IO_E;

/* Close and restore any files */

	for (i = STDIN_FILENO; i <= STDERR_FILENO; i++)
	{
	    if (sp->fp[i] != -1)
	    {
		S_close (i, TRUE);
		dup2 (sp->fp[i], i);
		S_close (sp->fp[i], TRUE);
	    }
	}
    }

    return rv;
}

/*
 * Create a Pipe
 */

int		openpipe ()
{
    register int	i;

    if ((i = S_open (TRUE, g_tempname (), O_PMASK, 0600)) < 0)
	print_error (nopipe);

    return i;
}

/*
 * Close a pipe
 */

void		closepipe (pv)
register int	pv;
{
    if (pv != -1)
	S_close (pv, TRUE);
}

/*
 * Write a character to STDERR
 */

void	S_putc (c)
int	c;
{
    write (STDERR_FILENO, (char *)&c, 1);
}

/*
 * Write a string to STDERR
 */

void	S_puts (s)
char	*s;
{
    write (STDERR_FILENO, s, strlen (s));
}

/*
 * Check for restricted shell
 */

bool	check_rsh (s)
char	*s;
{
    if (r_flag)
    {
	print_error ("%s: restricted\n", s);
	return TRUE;
    }

    return FALSE;
}

/*
 * Check to see if a file is a shell script.  If it is, return the file
 * handler for the file
 */

int	O_for_execute (path)
char	*path;
{
    int		i, end;
    char	local_path[FFNAME_MAX];

/* Work on a copy of the path */

    strcpy (local_path, path);

/* Try the file name and then with a .sh appended */

    for (end = 0; end < 2; end++)
    {
	if ((i = Check_Script (local_path)) >= 0)
	    return i;

	if (!end)
	    strcat (local_path, ".sh");
    }

    return -1;
}

/*
 * Check for shell script
 */

int	Check_Script (path)
char	*path;
{
    char	buf[5];
    int		i;

    if (((i = S_open (FALSE, path, O_RMASK)) >= 0) &&
	((read (i, buf, 6) == 5) && (strncmp (buf, "#!sh\n", 5) == 0)))
	return i;

    if (i != -1)
	S_close (i, TRUE);

    return -1;
}

/*
 * Convert slashes to backslashes for MSDOS
 */

void	Convert_Slashes (sp)
char	*sp;
{
    while (*sp)
    {
	if (*sp == '/')
	    *sp = '\\';

	++sp;
    }
}
