/* MS-DOS SHELL - History Processing
 *
 * MS-DOS SHELL - Copyright (c) 1990 Data Logic Limited
 *
 * This code is subject to the following copyright restrictions:
 *
 * 1.  Redistribution and use in source and binary forms are permitted
 *     provided that the above copyright notice is duplicated in the
 *     source form and the copyright notice in file sh6.c is displayed
 *     on entry to the program.
 *
 * 2.  The sources (or parts thereof) or objects generated from the sources
 *     (or parts of sources) cannot be sold under any circumstances.
 *
 *    $Header: sh9.c 1.1 90/01/26 17:25:19 MS_user Exp $
 *
 *    $Log:	sh9.c $
 * Revision 1.1  90/01/26  17:25:19  MS_user
 * Initial revision
 * 
 * 
 */

#include <sys/types.h>
#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <signal.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <setjmp.h>
#include <limits.h>
#include <dos.h>
#include <unistd.h>
#include "sh.h"

static bool	alpha_numeric (int);
static bool	function (int);
static bool	Process_History (int);
static bool	Scan_History (void);
static void	Redisplay_Line (void);
static void	Process_Stdin (void);
static void	Page_History (int);
static bool	UpDate_CLine (char *);
static bool	Re_start (char *);
static void	memrcpy (char *, char *, int);
static int	read_cursor_position (void);
static void	set_cursor_position (int);
static void	gen_cursor_position (void);

static bool	insert_mode = FALSE;
static char	*c_buffer_pos;		/* Position in command line	*/
static char	*end_buffer;		/* End of command line		*/
static int	s_cursor;		/* Start cursor position	*/
static int	c_history = -1;		/* Current entry		*/
static int	l_history = 0;		/* End of history array		*/
static int	M_length = -1;		/* Match length			*/
static char	l_buffer[LINE_MAX + 1];
static char	*No_prehistory   = "history: No previous commands\033[2K\n";
static char	*No_MatchHistory = "history: No history match found\033[2K\n";
static char	*No_posthistory  = "history: No more commands\033[2K\n";
static char	*History_2long   = "history: History line too long\033[2K\n";
static char	*H_TooLongI = "History file line too long - ignored (%d)\n";

/* Arrary of history Items */

static struct	cmd_history {
    int		number;
    char	*command;
} cmd_history[HISTORY_MAX];

/* Processing standard input */

int			Get_stdin (ap)
register IO_Args	*ap;
{
    int		coff = (int)ap->afpos;
    char	rv;

/* Is there anything in the input buffer.  If not, add the previous line to
 * the history buffer and get the next line
 */

    if (!coff)
	Process_Stdin ();			/* No - get input	*/

/* Get the next character */

    if ((rv = l_buffer[coff]) == NL)
    {
	l_buffer[coff] = 0;
	ap->afpos = 0L;
    }

/* Check for end of file */

    else if (rv == 0x1a)
    {
	l_buffer[coff] = 0;
	ap->afpos = 0L;
	rv = 0;
    }

    else
	ap->afpos++;

    return rv;
}

/* Input processing function */

static void	Process_Stdin ()
{
    int		i;
    char	*control = "^x";

/* Set to last history item */

    c_history = l_history - 1;

/* Process the input */

    while (TRUE)
    {
	c_buffer_pos = l_buffer;	/* Initialise			*/
	end_buffer = l_buffer;
	insert_mode = FALSE;
	M_length = -1;
	s_cursor = read_cursor_position ();

	while (((i = getch ()) != 0x1a) && (i != NL) && (i != '\r'))
	{

/* Re-position the line? */

	    if (((i) ? alpha_numeric (i) : function (getch ())))
		Redisplay_Line ();

/* Reposition the cursor */

	    gen_cursor_position ();
	}

/* Terminate the line */

	*end_buffer = 0;
	v1_putc (NL);

/* Line input - check for history */

	if ((*l_buffer == '!') && Process_History (0))
	{
	    v1a_puts (l_buffer);
	    break;
	}
	
	else if (*l_buffer != '!')
	    break;

/* Output prompt and try again */

	Re_start ((char *)NULL);
    }

    *end_buffer = (char)((i == '\r') ? NL : i);
}

/* Handler Alpha_numeric characters */

static bool	alpha_numeric (c)
int		c;
{
    bool	redisplay = FALSE;

/* Backspace processing */

    if (c == 0x08)
    {
	if (c_buffer_pos == l_buffer)
	{
	    v1_putc (0x07);		/* Ring bell			*/
	    return FALSE;
	}

/* Decrement current position */

	if ((c_buffer_pos--) == end_buffer)
	    --end_buffer;

	else
	    *c_buffer_pos = ' ';

        return TRUE;
    }

/* Normal character processing */

    if ((c_buffer_pos - l_buffer) == LINE_MAX)
    {
	v1_putc (0x07);			/* Ring bell			*/
	return FALSE;
    }

    else if (!insert_mode)
    {
	if (c_buffer_pos == end_buffer)
	    ++end_buffer;

	else if (iscntrl (*c_buffer_pos) || iscntrl (c))
	    redisplay = TRUE;

	*(c_buffer_pos++) = (char)c;

	if (redisplay || (c == '\t'))
	    return TRUE;

	if (iscntrl (c))
	{
	    v1_putc ('^');
	    c += '@';
	}

	v1_putc ((char)c);
	return FALSE;
    }

    else if ((end_buffer - l_buffer) == LINE_MAX)
    {
	v1_putc (0x07);			/* Ring bell - line full	*/
	return FALSE;
    }

    else
    {
	if (c_buffer_pos != end_buffer)
	    memrcpy (end_buffer + 1, end_buffer, end_buffer - c_buffer_pos + 1);

	++end_buffer;
	*(c_buffer_pos++) = (char)c;
	return TRUE;
    }
}

/* Process function keys */

static bool	function (fn)
int		fn;
{
    switch (fn)
    {
	case 'I':			/* Scan back command line	*/
	case 'Q':			/* Scan up command line		*/
	    if (M_length == -1)
		break;

	    Page_History ((fn == 'I') ? -1 : 1);
	    return TRUE;

	case 'H':			/* Previous command line	*/
	    Process_History (-1);
	    return TRUE;

	case 'P':			/* Next command line		*/
	    Process_History (1);
	    return TRUE;

	case 'K':			/* Cursor left			*/
	    if (c_buffer_pos != l_buffer)
		--c_buffer_pos;

	    else
		v1_putc (0x07);

	    return FALSE;

	case 'M':			/* Cursor right			*/
	    if (c_buffer_pos != end_buffer)
		++c_buffer_pos;

	    else
		v1_putc (0x07);

	    return FALSE;

	case 's':			/* Cursor left a word		*/
	    if (c_buffer_pos != l_buffer)
	    {
		--c_buffer_pos;		/* Reposition on previous char	*/

		while (isspace (*c_buffer_pos) && (c_buffer_pos != l_buffer))
		    --c_buffer_pos;

		while (!isspace (*c_buffer_pos) && (c_buffer_pos != l_buffer))
		    --c_buffer_pos;

		if (c_buffer_pos != l_buffer)
		    ++c_buffer_pos;
	    }

	    else
		v1_putc (0x07);

	    return FALSE;

	case 't':			/* Cursor right a word		*/
	    if (c_buffer_pos != end_buffer)
	    {

/* Skip to the end of the current word */

		while (!isspace (*c_buffer_pos) && (c_buffer_pos != end_buffer))
		    ++c_buffer_pos;

/* Skip over the white space */

		while (isspace (*c_buffer_pos) && (c_buffer_pos != end_buffer))
		    ++c_buffer_pos;
	    }

	    else
		v1_putc (0x07);

	    return FALSE;

	case 'G':			/* Cursor home			*/
	    c_buffer_pos = l_buffer;
	    return FALSE;

	case 'u':			/* Flush to end			*/
	    memset (c_buffer_pos, ' ', end_buffer - c_buffer_pos);
	    end_buffer = c_buffer_pos;
	    return TRUE;

	case 'O':			/* Cursor end of command	*/
	    if (*l_buffer == '!')
	    {
		*end_buffer = 0;
		Process_History (2);
		return TRUE;
	    }

	    c_buffer_pos = end_buffer;
	    return FALSE;

	case 'R':			/* Switch insert mode		*/
	    insert_mode = (insert_mode) ? FALSE : TRUE;
	    return FALSE;

	case 'S':			/* Delete character		*/
	    if (c_buffer_pos != end_buffer)
		memcpy (c_buffer_pos, c_buffer_pos + 1,
			end_buffer - c_buffer_pos);

	    if (end_buffer == l_buffer)
	    {
		v1_putc (0x07);
		return TRUE;
	    }

	    if (--end_buffer < c_buffer_pos)
		--c_buffer_pos;

   	    return TRUE;
    }

    v1_putc (0x07);
    return FALSE;
}

/* Read Cursor position */

static int	read_cursor_position ()
{
    union REGS	r;

    r.h.ah = 0x03;				/* Read cursor position	*/
    r.h.bh = 0;					/* Page zero		*/
    int86 (0x10, &r, &r);
    return (r.h.dh * 80) + r.h.dl;
}

/* Re-position the cursor */

static void	set_cursor_position (new)
int		new;
{
    union REGS	r;

    r.h.ah = 0x02;				/* Set new position	*/
    r.h.bh = 0;					/* Page zero		*/
    r.h.dh = (unsigned char)(new / 80);
    r.h.dl = (unsigned char)(new % 80);

/* Are we at the bottom of the page? */

    if (r.h.dh == 25)
    {
	r.h.dh = 24;
	s_cursor -= 80;
    }

    int86 (0x10, &r, &r);
}

/* Generate the new cursor position */

static void	gen_cursor_position ()
{
    char	*cp = l_buffer - 1;
    int		off = s_cursor;

/* Search to current position */

    while (++cp != c_buffer_pos)
    {
	if (*cp == '\t')
	    while ((++off) % 8);

	else if (iscntrl (*cp))
	    off += 2;

	else
	    ++off;
    }

/* Position the cursor */

    set_cursor_position (off);
}

/* Redisplay the current line */

static void	Redisplay_Line ()
{
    char	*control = "^x";
    char	*cp = l_buffer;
    int		off = s_cursor;

/* Reposition to start of line */

    set_cursor_position (s_cursor);

/* Output the line */

    while (cp != end_buffer)
    {
	if (*cp == '\t')
	{
	    do
	    {
		v1_putc (SP);
	    } while ((++off) % 8);
	}

	else if (iscntrl (*cp))
	{
	    control[1] = *cp + '@';
	    v1_puts (control);
	    off += 2;
	}

	else
	{
	    ++off;
	    v1_putc (*cp);
	}

	++cp;
    }

    v1_puts ("\033[2K");		/* clear to end of line	*/
}

/* Process history command
 *
 * -1: Previous command
 *  1: Next command
 *  0: Current command
 *  2: Current command with no options processing
 */

static bool	Process_History (direction)
int		direction;
{
    char	*optionals = null;

    c_buffer_pos = l_buffer;
    end_buffer = l_buffer;
    c_history += (direction == 2) ? 0 : direction;

    switch (direction)
    {
	case -1:			/* Move up one line		*/
	    if (c_history < 0)
	    {
		++c_history;
		return Re_start (No_prehistory);
	    }

	    break;

	case 1:				/* Move to next history line	*/
	    if (c_history >= l_history)
	    {
		--c_history;
		return Re_start (No_posthistory);
	    }

	    break;

	case 0:				/* Check out l_buffer		*/
	    optionals = l_buffer;	/* Are there any additions to	*/
					/* the history line		*/

/* Find the end of the first part */

	    while (!isspace (*optionals) && *optionals)
		++optionals;

/* Terminate the history command */

	    if (*optionals)
		*(optionals++) = 0;

/* Find the end of the space separator part which gives the start of the
 * optionals
 */

	    while (isspace (*optionals))
		++optionals;
	    

/* Copy selected item into line buffer */

	case 2:
	    M_length = strlen (l_buffer) - 1;
	    if (!Scan_History ())
		return FALSE;

	    break;
    }

    return UpDate_CLine (optionals);
}

/* Ok c_history points to the new line.  Move optionals after history
 * and the copy in history and add a space
 */

static bool	UpDate_CLine (optionals)
char		*optionals;
{
    int		opt_len;

    end_buffer = &l_buffer[strlen (cmd_history[c_history].command)];

    if ((end_buffer - l_buffer + (opt_len = strlen (optionals)) + 1) >= LINE_MAX)
	return Re_start (History_2long);

    if (end_buffer > optionals)
	memrcpy (end_buffer + 1 + opt_len, optionals + opt_len, opt_len + 1);
    
    else
	strcpy (end_buffer + 1, optionals);

    strcpy (l_buffer, cmd_history[c_history].command);

    if (opt_len)
	*end_buffer = ' ';

    end_buffer = &l_buffer[strlen (l_buffer)];
    return TRUE;
}

/* Scan the line buffer for a history match */

static bool	Scan_History ()
{
    char	*cp = &l_buffer[1];
    int		c_len = strlen (cp);
    char	*ep;
    int		i = (int)strtol (cp, &ep, 10);

/* Get the previous command ? (single !) */

    if (c_len == 0)
    {
	if (c_history < 0)
	{
	    M_length = -1;
	    return Re_start (No_prehistory);
	}

	return TRUE;
    }

/* Request for special history number item.  Check History file empty */

    if (l_history == 0)
    {
	M_length = -1;
	return Re_start (No_MatchHistory);
    }

/* Check for number */

    if (!*ep)
    {
	M_length = -1;

	for (c_history = l_history - 1;
	    (cmd_history[c_history].number != i) && (c_history >= 0);
	    --c_history);
    }

/* No - scan for a match */

    else
    {
	for (c_history = l_history - 1;
	    (strncmp (cp, cmd_history[c_history].command, c_len) != 0)
	     && (c_history >= 0);
	    --c_history);
    }

/* Anything found ? */

    if (c_history == -1)
    {
	c_history = l_history - 1;
	return Re_start (No_MatchHistory);
    }

    return TRUE;
}

/* Scan back or forward from current history */

static void	Page_History (direction)
int		direction;
{
    c_buffer_pos = l_buffer;
    end_buffer = l_buffer;

    if (l_history == 0)
    {
	M_length = -1;
	Re_start (No_MatchHistory);
	return;
    }

/* scan for a match */

    while (((c_history += direction) >= 0) && (c_history != l_history) &&
	   (strncmp (l_buffer, cmd_history[c_history].command, M_length) != 0));

/* Anything found ? */

    if ((c_history < 0) || (c_history == l_history))
    {
	c_history = l_history - 1;
	Re_start (No_MatchHistory);
    }

    else
	UpDate_CLine (null);
}

/* Load history file */

void	Load_History ()
{
    FILE		*fp;
    char		*cp;
    int			i = 0;
    Var_List		*lset;

/* Initialise history array */

    memset (cmd_history, 0, sizeof (struct cmd_history) * HISTORY_MAX);
    c_history = -1;			/* Current entry		*/
    l_history = 0;			/* End of history array		*/

    if ((lset = lookup (history_file, TRUE))->value == null)
	setval (lset, strcat (strcpy (l_buffer, lookup (home, FALSE)->value),
			      "history.sh"));

    if (!History_Enabled || ((fp = fopen (lset->value, "rt")) == (FILE *)NULL))
	return;

/* Read in file */

    while (fgets (l_buffer, LINE_MAX, fp) != (char *)NULL)
    {
	++i;

	if ((cp = strchr (l_buffer, NL)) == (char *)NULL)
	    print_warn (H_TooLongI, i);

	else
	{
	    *cp = 0;
	    Add_History (TRUE);
	}
    }

    fclose (fp);
}

/* Add entry to history file */

void	Add_History (past)
bool	past;				/* Past history?	*/
{
    int			i;

    if ((!History_Enabled) || (strlen (l_buffer) == 0))
	return;

/* If adding past history, decrement all numbers previous */

    if ((past) && l_history)
    {
	for (i = 0; i < l_history; i++)
	    --(cmd_history[i].number);
    }

/* If the array is full, remove the last item */

    if (l_history == HISTORY_MAX)
    {
	if (cmd_history[0].command != null)
	    DELETE (cmd_history[0].command);

	--l_history;
	memcpy (&cmd_history[0], &cmd_history[1],
		sizeof (struct cmd_history) * (HISTORY_MAX - 1));
    }

/* If there are any items in the array */

    c_history = l_history;
    Current_Event = (l_history) ? cmd_history[l_history - 1].number + 1 : 0;
    cmd_history[l_history].number = Current_Event;

/* Save the string */

    cmd_history[l_history++].command = strsave (l_buffer, 0);
}

/* Print history */

void	Display_History ()
{
    int			i;
    struct cmd_history	*cp = cmd_history;

    if (!l_history)
	return;

    for (i = 0; i < l_history; ++cp, ++i)
    {
	v1printf ("%5d: ", cp->number);
	v1a_puts (cp->command);
    }
}

/* Dump history to file */

void	Dump_History ()
{
    int			i;
    struct cmd_history	*cp = cmd_history;
    FILE		*fp;

    if (!History_Enabled ||
	((fp = fopen (lookup (history_file, FALSE)->value, "wt")) ==
	 (FILE *)NULL))
	return;

    for (i = 0; i < l_history; ++cp, ++i)
    {
	fputs (cp->command, fp);
	fputc (NL, fp);
    }

    fclose (fp);
}

/* Clear out history */

void	Clear_History ()
{
    int			i;
    struct cmd_history	*cp = cmd_history;

    for (i = 0; i < l_history; ++cp, ++i)
    {
	if (cp->command != null)
	    DELETE (cp->command);
    }

    memset (cmd_history, 0, sizeof (struct cmd_history) * HISTORY_MAX);

    c_history = -1;			/* Current entry		*/
    l_history = 0;			/* End of history array		*/
    Current_Event = 0;
}

/* Output warning message and prompt */

static bool	Re_start (cp)
char		*cp;
{
    if (cp != (char *)NULL)
	print_warn (cp);

    put_prompt (last_prompt);

/* Re-initialise */

    c_buffer_pos = l_buffer;
    end_buffer = l_buffer;
    s_cursor = read_cursor_position ();

    return FALSE;
}

/* Copy backwards */

static void	memrcpy (sp1, sp, cnt)
char		*sp1;
char		*sp;
int		cnt;
{
    while (cnt--)
	*(sp1--) =  *(sp--);
}
